/*
 * Khoros: $Id$
 */

#if !defined(__lint) && !defined(__CODECENTER__)
static char rcsid[] = "Khoros: $Id$";
#endif

/*
 * $Log$
 */

/*
 * Copyright (C) 1993, 1994, 1995, Khoral Research, Inc., ("KRI").
 * All rights reserved.  See $BOOTSTRAP/repos/license/License or run klicense.
 */

/***************************************************************************
  Contents

  Private:  KMPI_Comm_init
  Public :  KMPI_Init
            KMPI_Finalize
  **************************************************************************/

/***************************************************************************
  Include files

  I define KMPILIBRARY so that MPI_INIT and MPI_FINALIZE don't get
  defined to be the routines in this file.
  **************************************************************************/
#define  KMPILIBRARY
#include "internals.h"
#include "stdio.h"
#include "malloc.h"

/***************************************************************************
  Local Variables
  
  CONNECTION      - macro used to access c_table as though it's a 2-d array
  MAX_CONNECTIONS - the maximum number of connections we can have
  app_table       - list of the application numbers for each process
  c_table         - list of input/output communicator connections
  **************************************************************************/
#define CONNECTION(i,j) (c_table[i*2+j])
#define MAX_CONNECTIONS  100
static int     *app_table, c_table[MAX_CONNECTIONS];
static int      c_table_size = 0, app_table_size = 0;


/***************************************************************************
  Global Variables
  
  KMPI_COMM_WORLD - The MPI world after subdivision into work units
  kmpi_in         - handle with input communicators
  kmpi_out        - handle with output communicators
  **************************************************************************/
MPI_Comm KMPI_COMM_WORLD = (MPI_Comm)0;
KMPI_Handle kmpi_in, kmpi_out;


/***************************************************************************
  KMPI_Comm_init

  kin  - Khoros MPI input communicator handle
  kout - Khoros MPI output communicator handle

  This function reads information from the connection table,
  creates the input (max# is 1) and output communicators, and stores
  these in the Khoros MPI input/output handles.
  **************************************************************************/
static int KMPI_Comm_init( KMPI_Handle *kin, KMPI_Handle *kout)
{
    int rank,size,i,j;
    int c_kin=0, c_kout=0;
    
    /* Initialize the KMPI handles */
    kin->num_comms  = 0;
    kin->comm       = (MPI_Comm *)0;
    kout->num_comms = 0;
    kout->comm      = (MPI_Comm *)0;
    
    /* create the local communicator */
    MPI_Comm_rank  ( MPI_COMM_WORLD, &rank );
    MPI_Comm_size  ( MPI_COMM_WORLD, &size );
    MPI_Comm_split ( MPI_COMM_WORLD, app_table[rank], rank, &KMPI_COMM_WORLD );
    
    /* count the number of ins and outs */
    for (i=0; i<c_table_size; i++) {
        if (CONNECTION(i,0) == app_table[rank])
            kout->num_comms++;
        if (CONNECTION(i,1) == app_table[rank])
            kin->num_comms++;
    }

    /* If we have more input communicators than 1, then there is an error */
    /* Thom disagrees.  Something like matrix addition would have 2 inputs */
/*
    if (kin->num_comms < 0 || kin->num_comms > 1) {
        fprintf(stderr, "ERROR: Illegal number of input communicators: %d\n",
                kin->num_comms);
        MPI_Abort(MPI_COMM_WORLD, MPI_ERR_OTHER);
    }
*/
    
    /* create the arrays of ins and outs */
    if (kin->num_comms)
        kin->comm  = (MPI_Comm *)malloc(sizeof(MPI_Comm) * kin->num_comms);
    if (kout->num_comms)
        kout->comm = (MPI_Comm *)malloc(sizeof(MPI_Comm) * kout->num_comms);

    /* Create the intercommunicators */
    for (i=0; i<c_table_size; i++) {
        if ( app_table[rank] == CONNECTION(i,0) ) {
            int remote_leader;
            int remote_app = CONNECTION(i,1);

            if (remote_app == -1) {
                free (kout->comm[c_kout]);
                kout->comm[c_kout++] = NULL;
            } else {
                for (j=0; j<app_table_size; j++)
                    if (app_table[j] == remote_app) {
                        remote_leader = j;
                        break;
                    }
            
                MPI_Intercomm_create ( KMPI_COMM_WORLD, 0, MPI_COMM_WORLD,
                                   remote_leader,  0, &kout->comm[c_kout++] );
            }
        }
        else if ( app_table[rank] == CONNECTION(i,1) ) {
            int remote_leader;
            int remote_app = CONNECTION(i,0);
            
            if (remote_app == -1) {
                free (kout->comm[c_kout]);
                kout->comm[c_kout++] = NULL;
            } else {
                for (j=0; j<app_table_size; j++)
                    if (app_table[j] == remote_app) {
                        remote_leader = j;
                        break;
                    }

                MPI_Intercomm_create ( KMPI_COMM_WORLD, 0, MPI_COMM_WORLD,
                                      remote_leader, 0, &kin->comm[c_kin++] );
            }
        }
    }
}


/***************************************************************************
  KMPI_Init

  Initializes the the Khoros MPI Environment by calling MPI_Init,
  reading the application and connection files, then creating the
  application and connection tables for use in building input/output
  communicators.
  **************************************************************************/
int KMPI_Init (int *argc, char ***argv)
{
    int      size, rank;
    FILE    *pgfile, *cfile;
    char     buf[1024];
    char    *bufp;
    char     buf1[1024], buf2[1024];
    int      i;
    int      mpi_errno;
    
    mpi_errno = MPI_Init(argc,argv);
    MPI_Comm_size (MPI_COMM_WORLD, &size);
    MPI_Comm_rank (MPI_COMM_WORLD, &rank);

    /* Create the app_table */
    app_table_size = size;
    app_table = (int *)malloc(app_table_size * sizeof(int));
    
    /* If I'm the kmpihost */
    if (rank == 0) {

        /* Read the connection file */
        if ((cfile = fopen((*argv)[2], "r")) == (FILE *)0) {
            fprintf(stderr, "ERROR: Couldn't open connection file: %s\n", (*argv)[2]);
            MPI_Abort(MPI_COMM_WORLD, MPI_ERR_OTHER);
        }

        /* Load the connection file into the connection table */
        while (fgets(buf, sizeof(buf), cfile) != (char *)0) {
            sscanf(buf, "%d %d",
                   &CONNECTION(c_table_size,0), &CONNECTION(c_table_size,1));
            c_table_size++;
            if (c_table_size > MAX_CONNECTIONS) {
                fprintf(stderr, "ERROR: Connection file size must be less than %d\n",
                        MAX_CONNECTIONS);
                MPI_Abort(MPI_COMM_WORLD, MPI_ERR_OTHER);
            }
        }
        
        /* Read the pg file and create app table */
        if ((pgfile = fopen((*argv)[1], "r")) == (FILE *)0) {
            fprintf(stderr, "ERROR: Couldn't open process file: %s\n", (*argv)[1]);
            MPI_Abort(MPI_COMM_WORLD, MPI_ERR_OTHER);
        }

        app_table[i] = 0; i = 1; 
        while (fgets(buf, sizeof(buf), pgfile) != (char *)0) {
            sscanf(buf, "%s %s %d",
                   buf1, buf2,
                   &app_table[i]);
            i++;
        }
    }    

    /* Broadcast the app_table and c_table */
    MPI_Bcast(&c_table_size, 1, MPI_INT, 0, MPI_COMM_WORLD);
    MPI_Bcast(app_table, app_table_size, MPI_INT, 0, MPI_COMM_WORLD);
    MPI_Bcast(c_table,   c_table_size*2, MPI_INT, 0, MPI_COMM_WORLD);

    /* Create the communicators */
    KMPI_Comm_init( &kmpi_in, &kmpi_out );

    /* Return what MPI_Init gave us */
    return (mpi_errno);
}


/***************************************************************************
  KMPI_Finalize

  Destroys the data structures built for KMPI, then destroys the MPI
  environment.
  **************************************************************************/
int KMPI_Finalize()
{
    int i;
    
    /* Free the tables */
    free(app_table); free(c_table);

    /* free the communicators */
    MPI_Comm_free ( &KMPI_COMM_WORLD );
    for (i=0;i< kmpi_in.num_comms;i++)
        MPI_Comm_free( &kmpi_in.comm[i] );
    for (i=0;i<kmpi_out.num_comms;i++)
        MPI_Comm_free( &kmpi_out.comm[i] );
    free(kmpi_in.comm); free(kmpi_out.comm);
    
    return (MPI_Finalize());
}


